#!/usr/bin/python3
from PyQt5 import QtWidgets, QtCore, QtMultimedia

try:
    import qoob.backend.parser as parser
except ImportError:
    import backend.parser as parser

REPEAT_ALL = 1
REPEAT_SINGLE = 2


class MediaPlayer(QtMultimedia.QMediaPlayer):
    currentMediaChanged = QtCore.pyqtSignal(str)
    setItemColor = QtCore.pyqtSignal(str)
    setSliderRange = QtCore.pyqtSignal(int, int)
    setSliderValue = QtCore.pyqtSignal(int)
    setPlayPauseIcon = QtCore.pyqtSignal(object)
    setShuffleIcon = QtCore.pyqtSignal(object)
    setRepeatIcon = QtCore.pyqtSignal(object)
    setStatusMessage = QtCore.pyqtSignal(str, str)

    def __init__(self, parent):
        super().__init__()
        self.parent = parent
        self.icon = parent.icon
        self.log = parent.logger.new("Player")
        self.preferences = parent.preferences
        self.mediaStatusChanged.connect(self._mediaChangedEvent)
        self.durationChanged.connect(self._updateDuration)
        self.positionChanged.connect(self._updateSlider)
        self.seekableChanged.connect(self._seekableEvent)
        self.random = QtCore.QRandomGenerator()
        self.playLater = QtCore.QTimer(singleShot=True, interval=500)
        self.playLater.timeout.connect(self.nextEvent)

        self.setVolume(self.preferences.get("state", "volume"))
        self.lastPlayed = {}
        self.lastState = None
        self.lastItem = None
        self.mediaLength = "?"
        self.path = ""

        self.statusTimer = QtCore.QTimer(interval=500)
        self.statusTimer.timeout.connect(self._updateStatus)
        self.statusTimer.start()

    def _currentItem(self):
        return self.parent.tabWidget.current.currentItem()

    def _currentTab(self):
        index = self.parent.tabWidget.currentIndex()
        tabName = self.parent.tabWidget.tabText(index)
        tabName = tabName.lstrip("&")  # KDE bug fix (auto append of &)
        return tabName

    def _currentView(self):
        return self.parent.tabWidget.current

    def _mediaChangedEvent(self, event):
        tab = self._currentTab()
        if not tab in self.lastPlayed:
            self.lastPlayed[tab] = []

        currentRow = self._currentView().currentIndex().row()
        if event == QtMultimedia.QMediaPlayer.EndOfMedia:
            self.setPosition(0)
            self.nextEvent()
        elif event == QtMultimedia.QMediaPlayer.InvalidMedia:
            self.lastPlayed[tab].append(currentRow)
            self.setItemColor.emit("red")
            self.setStatusMessage.emit(f"! {self.path}", "Invalid media")
            self.playLater.start()  # Cannot call nextEvent directly (too fast)
        elif event == QtMultimedia.QMediaPlayer.BufferedMedia:
            if not self.repeat == REPEAT_SINGLE:
                self.lastPlayed[tab].append(currentRow)
                self.log.info(f"'{self._currentTab()}' playlist: {self.lastPlayed[tab]}")

    def _seekableEvent(self):
        if self.resumePlayback and self.isSeekable():
            self.resumePlayback = False
            self.setPosition(self.preferences.get("state", "playback position"))

    def _setCurrentItem(self, item):
        # Slots/signals too slow, sync required
        self._currentView().setCurrentItem(item)

    def _nextPlay(self):
        next = self._currentView().itemBelow(self._currentItem())
        self._setCurrentItem(next)
        self.activateSelection()

    def _nextRepeat(self):
        next = self._currentView().topLevelItem(0)
        self._setCurrentItem(next)
        self.activateSelection()

    def _nextShuffle(self):
        currentRow = self._currentView().currentIndex().row()
        itemCount = self._currentView().topLevelItemCount()
        tab = self._currentTab()
        self.lastPlayed.setdefault(tab, [])

        if len(self.lastPlayed[tab]) == itemCount:
            self.lastPlayed[tab] = []
            if self.repeat == REPEAT_ALL:
                self.nextEvent()
            else:
                self.stopEvent()
        else:
            next = self._shuffleTrack(itemCount)
            self._setCurrentItem(next)
            self.activateSelection()

    def _shuffleTrack(self, count):
        tab = self._currentTab()
        next = self.random.bounded(count)
        while next in self.lastPlayed[tab]:
            next = self.random.bounded(count)
        return self._currentView().topLevelItem(next)

    def _updateDuration(self, duration):
        s = duration / 1000
        m, s = divmod(s, 60)
        h, m = divmod(m, 60)
        self.mediaLength = "%02d:%02d:%02d" % (h, m, s)
        self.setSliderRange.emit(0, self.duration())

    def _updateSlider(self, progress):
        if not self.parent.ui.playbackSlider.isSliderDown():
            self.setSliderValue.emit(progress)

    def _updateStatus(self):
        if self.duration() > -1:
            right, left = None, None

            if not self.state() == self.lastState:
                if self.state() == QtMultimedia.QMediaPlayer.PlayingState:
                    left = "Now playing: " + self.tags["title"] + " (" + self.tags["artist"] + ")"
                elif self.state() == QtMultimedia.QMediaPlayer.PausedState:
                    left = "Paused"
                elif self.state() == QtMultimedia.QMediaPlayer.StoppedState:
                    left = "Stopped"

            if self.state() == QtMultimedia.QMediaPlayer.PlayingState and not self.mediaLength == "?":
                s = self.parent.ui.playbackSlider.value() / 1000
                m, s = divmod(s, 60)
                h, m = divmod(m, 60)
                elapsed = "%02d:%02d:%02d" % (h, m, s)
                right = f"    {elapsed} / {self.mediaLength}"

            self.lastState = self.state()
            self.setStatusMessage.emit(left, right)

    def activateSelection(self):
        self.resumePlayback = False
        if self.lastItem:
            self.lastItem.setColor("none")

        if self._currentItem():
            self._currentItem().setColor("green")
            self.setCurrentMedia(self._currentItem().text(5))
            self.lastItem = self._currentItem()

        elif self._currentView().topLevelItemCount() > 0:
            first = self._currentView().topLevelItem(0)
            self._currentView().setCurrentItem(first)
            self._setCurrentItem(first)
            self.activateSelection()

    def clearTabPlaylist(self, tab):
        if tab in self.lastPlayed:
            del self.lastPlayed[tab]

    def clearViewerPlaylist(self):
        self.lastPlayed["Library viewer"] = []

    def nextEvent(self, forced=False):
        currentRow = self._currentView().currentIndex().row()
        itemCount = self._currentView().topLevelItemCount()
        if self._currentItem():
            if self.repeat == REPEAT_SINGLE:
                self.activateSelection()
            elif self.shuffle:
                self._nextShuffle()
            elif currentRow == itemCount - 1:
                if self.repeat == REPEAT_ALL or (forced and self.state() == QtMultimedia.QMediaPlayer.StoppedState):
                    tab = self._currentTab()
                    self.lastPlayed[tab] = []
                    self._nextRepeat()
                else:
                    self.stopEvent()
            else:
                self._nextPlay()
        elif self.shuffle:
            self._nextShuffle()
        else:
            self.activateSelection()

    def pauseEvent(self):
        self.playLater.stop()
        self.pause()
        self.setPlayPauseIcon.emit(self.icon["play"])

    def playEvent(self):
        self.playLater.stop()
        if self.state() == QtMultimedia.QMediaPlayer.PausedState:
            self.play()
            self.setPlayPauseIcon.emit(self.icon["pause"])
        elif self.state() == QtMultimedia.QMediaPlayer.StoppedState:
            self.activateSelection()

    def playPauseEvent(self):
        self.playLater.stop()
        if self.state() == QtMultimedia.QMediaPlayer.PlayingState:
            self.pauseEvent()
        else:
            self.playEvent()

    def previousEvent(self):
        self.playLater.stop()
        if self._currentItem():
            tab = self._currentTab()
            if self.lastPlayed[tab]:
                self.lastPlayed[tab].pop()
            if self.lastPlayed[tab]:
                previous = self.lastPlayed[tab].pop()
                previous = self._currentView().topLevelItem(previous)
            else:
                previous = self._currentView().itemAbove(self._currentItem())
                if not previous:
                    previous = self._currentView().topLevelItem(0)
            self._setCurrentItem(previous)
        self.activateSelection()

    def setCurrentMedia(self, path):
        self.path = path
        self.media = QtMultimedia.QMediaContent(QtCore.QUrl.fromLocalFile(self.path))
        self.setMedia(self.media)
        self.play()
        self.setPlayPauseIcon.emit(self.icon["pause"])
        self.tags = self.parent.parser.header(self.path)
        self.currentMediaChanged.emit(path)
        self.lastState = None
        if "error" in self.tags:
            self.setItemColor.emit("yellow")
            self.setStatusMessage.emit(self.path, self.tags["error"])

    def shuffleEvent(self, event=None, enable=None):
        self.shuffle = not self.shuffle if enable is None else enable
        if self.shuffle:
            self.setShuffleIcon.emit(self.icon["shuffle_on"])
        else:
            self.setShuffleIcon.emit(self.icon["shuffle_off"])

    def repeatEvent(self, event=None, state=None):
        self.repeat = self.repeat + 1 if state is None else state
        if self.repeat > 2:
            self.repeat = 0
        icon = {0: "repeat_off", 1: "repeat_all", 2: "repeat_single"}
        stateIcon = icon[self.repeat]
        self.setRepeatIcon.emit(self.icon[stateIcon])

    def skip(self, seconds):
        self.setPosition(self.position() + seconds*1000)

    def stopEvent(self):
        tab = self._currentTab()
        self.stop()
        self.playLater.stop()
        self.lastPlayed[tab] = []
        self.setSliderValue.emit(0)
        self.setPlayPauseIcon.emit(self.icon["play"])



#!/usr/bin/python3
import hashlib
import json
import os
import time
from collections import OrderedDict
from PyQt5 import QtWidgets, QtCore, QtGui

try:
    import qoob.backend.parser as parser
    import qoob.backend.tools as tools
    import qoob.frontend.playlist as playlist
except ImportError:
    import backend.parser as parser
    import backend.tools as tools
    import frontend.playlist as playlist


class LibraryTree(QtWidgets.QTreeWidget):
    selectEvent = QtCore.pyqtSignal(object)

    def __init__(self, parent):
        super().__init__()
        self.parent = parent
        self.preferences = parent.preferences
        self.setAlternatingRowColors(True)
        self.header().setVisible(False)
        self.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        self.setIndentation(10)
        self.itemSelectionChanged.connect(self._selectItem)

    def _selectItem(self):
        if self.selectedItems():
            self.selectEvent.emit(self.selectedItems())

    def addAlbum(self, node, path):
        name = os.path.splitext(os.path.basename(path))[0]
        item = QtWidgets.QTreeWidgetItem()
        item.setText(0, name)
        item.setText(1, path)
        try:
            node.addChild(item)
        except RuntimeError:
            pass

    def addAllMusic(self):
        item = QtWidgets.QTreeWidgetItem()
        item.setText(0, "All Music")
        item.setText(1, "__all__")
        self.insertTopLevelItem(0, item)

    def addArtist(self, node):
        self.addTopLevelItem(node)
        self.sortItems(0, QtCore.Qt.AscendingOrder)

    @QtCore.pyqtSlot()
    def expand(self):
        if self.preferences.get("viewer", "expand library"):
            count = self.topLevelItemCount()
            for item in range(self.topLevelItemCount()):
                count += self.topLevelItem(item).childCount()
            if (count * 12) <= self.height():  # Where 12px is the height of one item
                self.expandAll()
        if not self.parent.ui.libraryFilterLine.text():
            self.addAllMusic()


class LibraryWorker(QtCore.QObject):
    setStatusMessage = QtCore.pyqtSignal(str)
    addArtist = QtCore.pyqtSignal(object)
    state = QtCore.pyqtSignal(bool)
    expand = QtCore.pyqtSignal()

    def __init__(self, parent):
        super().__init__()
        self.parent = parent
        self.preferences = parent.preferences
        self.libraryView = parent.libraryView
        self.log = parent.logger.new("Library")
        self.cacheTree = tools.Database("cache_tree")
        self.cacheView = tools.Database("cache_view")
        self.selection = []
        self.siftAborted = False
        self.sift = ""

    def _buildTree(self, paths):
        tree = {}
        musicDb = self.preferences.get("general", "music database")
        for p in paths:
            if self._siftChanged():
                self.siftAborted = True
                break
            albumPath = os.path.abspath(os.path.join(p, os.pardir))
            if albumPath not in musicDb:
                artistPath = os.path.abspath(os.path.join(albumPath, os.pardir))
                if artistPath in musicDb:
                    artistPath = albumPath

                artist, album = self._tagFromPath(p)
                if artist not in tree:
                    tree[artist] = {"path": artistPath, "albums": {}}
                tree[artist]["albums"][album] = albumPath
        return tree

    def _cacheViewSave(self, available):
        self.cacheView.db[available] = []
        for item in range(self.libraryView.topLevelItemCount()):
            item = self.libraryView.topLevelItem(item)
            self.cacheView.db[available].append(item.text(5))
        self.cacheView.save()
        self.log.info(f"View cache saved for 'All Music'")

    def _cacheTreeRestore(self):
        self.setStatusMessage.emit("Loading library tree cache ...")
        for artist in self.cacheTree.db:
            if self._siftChanged(): break
            if self._isPathValid(artist):
                node = self._item(artist)
                for album in self.cacheTree.db[artist]:
                    if self._siftChanged(): break
                    if self._isPathValid(album):
                        album = self._item(album)
                        node.addChild(album)
                self.addArtist.emit(node)
        self.state.emit(False)

    def _fetchAllMusic(self):
        self.setStatusMessage.emit("Loading of 'All Music' view cache ...")
        paths = self.preferences.get("general", "music database")

        available = ""
        for p in paths:
            if os.path.isdir(p):
                available += f"%{p}%"

        if not self.parent.ui.libraryFilterLine.text():
            started = time.time()
            if available in self.cacheView.db:
                tags = self._fetchTag(self.cacheView.db[available])
                self._fillLibraryView(tags)
                elapsed = int((time.time() - started)*1000)
                self.log.info(f"View loaded from cache after {elapsed}ms, {len(tags)} items")
                self.setStatusMessage.emit(f"View loaded from cache after {elapsed}ms, {len(tags)} items")
            else:
                self._fetchView(paths)
                self._cacheViewSave(available)
        else:
            self._fetchView(paths)

    def _fetchTag(self, files):
        tags = {}
        for path in files:
            if self._siftChanged(): break
            if self._selectionChanged(): break
            if parser.allowedType(path):
                tags[path] = self.parent.parser.header(path)
        return tags

    def _fetchTree(self):
        if self.cacheTree.db:
            started = time.time()
            self._cacheTreeRestore()
            elapsed = int((time.time() - started)*1000)
            self.log.info(f"Library tree loaded from cache after {elapsed}ms")
            self.setStatusMessage.emit(f"Library tree loaded from cache after {elapsed}ms")
        else:
            QtCore.QMetaObject.invokeMethod(self.parent.parser, "scanAll", QtCore.Qt.QueuedConnection)

    def _fetchView(self, paths):
        tags = {}
        for p in paths:
            started = time.time()
            tags.update(self._selectionScan(p))
        tags = self._sortRoutine(tags)
        self._fillLibraryView(tags)
        elapsed = int((time.time() - started)*1000)
        itemsCount = self.libraryView.topLevelItemCount()
        self.log.info(f"View loaded after {elapsed}ms, {itemsCount} items")
        self.setStatusMessage.emit(f"View loaded after {elapsed}ms, {itemsCount} items")

    def _fillLibraryTree(self, tree):
        for artist in sorted(tree, key=str.lower):
            if self._siftChanged():
                self.siftAborted = True
                break
            artistPath = tree[artist]["path"]
            item = QtWidgets.QTreeWidgetItem()
            for album in sorted(tree[artist]["albums"], key=str.lower):
                albumPath = tree[artist]["albums"][album]
                if not albumPath == artistPath:
                    subItem = QtWidgets.QTreeWidgetItem()
                    subItem.setText(0, album)
                    subItem.setText(1, albumPath)
                    item.addChild(subItem)
            item.setText(0, artist)
            item.setText(1, artistPath)
            self.addArtist.emit(item)

    def _fillLibraryView(self, metadata):
        for path in metadata:
            if self._siftChanged(): break
            if self._selectionChanged(): break
            item = playlist.TreeItem()
            item.setText(0, metadata[path]["artist"])
            item.setText(1, metadata[path]["album"])
            item.setText(2, metadata[path]["track"])
            item.setText(3, metadata[path]["title"])
            item.setText(4, metadata[path]["duration"])
            item.setText(5, path)
            self.libraryView.addTopLevelItem(item)  ## QBasicTimer error?

    def _item(self, path):
        name = os.path.basename(path)
        item = QtWidgets.QTreeWidgetItem()
        item.setText(0, name)
        item.setText(1, path)
        return item

    def _isPathValid(self, path):
        for p in self.preferences.get("general", "music database"):
            if os.path.isdir(p) and path.startswith(f"{p}/"):
                return True
        return False

    def _selectionChanged(self):
        selection = self.parent.library.selectedItems()
        if selection:
            return not self.selection == selection
        return False

    def _selectionScan(self, path):
        sifted = set()
        for root, subfolder, files in os.walk(path):
            if self._siftChanged(): break
            if self._selectionChanged(): break
            for f in files:
                p = os.path.join(root, f)
                if self._siftView(f, p):
                    sifted.add(p)
        if os.path.isfile(path):
            if self._siftView(path):
                sifted.add(path)
        return self._fetchTag(sifted)

    def _siftChanged(self):
        return not self.sift == self.parent.ui.libraryFilterLine.text()

    def _siftTree(self, metadata, sift):
        result = set()
        for f in metadata:
            if self._siftChanged():
                self.siftAborted = True
                break

            if self._isPathValid(f):
                value = metadata[f]["title"].lower()
                if value.count(sift):
                    result.add(f)
        return result

    def _siftView(self, filename, path):
        metadata = self.parent.parser.metadata.db
        if path in metadata:
            tag = metadata[path]["title"]
            if tag.lower().count(self.sift.lower()):
                return True
        return False

    def _sortRoutine(self, metadata):
        direction = {"(ascending)": False, "(descending)": True}
        for action in self.preferences.get("viewer", "sorting routine"):
            if self._siftChanged(): return {}
            if self._selectionChanged(): return {}
            action = action.lower().split()
            try:
                column = action[0]
                reverse = direction[action[1]]
                metadata = self._sortMetadata(metadata, column, reverse=reverse)
            except KeyError:
                pass  # Legacy fix for "Disable sorting" key
        return metadata

    def _sortMetadata(self, db, column, reverse):
        if column == "track" or column == "duration":
            return OrderedDict(sorted(db.items(), key=lambda item:(self._toInt(item[1][column])), reverse=reverse))
        else:
            return OrderedDict(sorted(db.items(), key=lambda item:(item[1][column]), reverse=reverse))

    def _tagFromPath(self, path):
        for p in self.preferences.get("general", "music database"):
            if path.startswith(f"{p}/"):
                path = path[len(p)+1:]
        path = path.split("/")
        artist = path[0]
        album = path[1] if len(path) > 1 else None
        return (artist, album)

    def _toInt(self, string):
        try:
            if string.count(":"):
                string = string.split(":")
                return int(string[0])*3600 + int(string[1])*60 + int(string[2])
            else:
                return int(string)
        except ValueError:
            return 0

    @QtCore.pyqtSlot()
    def clearViewCache(self):
        self.cacheView.db = {}
        self.cacheView.save()
        self.log.info("Cleared view cache")

    @QtCore.pyqtSlot(str)
    def filterEvent(self, sift):
        started = time.time()
        QtCore.QMetaObject.invokeMethod(self.libraryView, "clear", QtCore.Qt.BlockingQueuedConnection)
        QtCore.QMetaObject.invokeMethod(self.parent.library, "clear", QtCore.Qt.BlockingQueuedConnection)
        self.sift = sift
        self.siftAborted = False

        if sift:
            # Search pattern in titles / filenames
            self.setStatusMessage.emit(f"Searching for '{sift}' ...")
            metadata = self.parent.parser.metadata.db
            sifted = self._siftTree(metadata, sift)
            tree = self._buildTree(sifted)
            tags = {}
            for artist in tree:
                if self._siftChanged():
                    self.siftAborted = True
                    break
                path = tree[artist]["path"]
                tags.update(self._selectionScan(path))
            tags = self._sortRoutine(tags)
            self._fillLibraryView(tags)
            self._fillLibraryTree(tree)

            self.expand.emit()
            elapsed = int((time.time() - started)*1000)
            if not self.siftAborted:
                self.log.info(f"Search of '{sift}' done after {elapsed}ms, {len(tags)} results")
                self.setStatusMessage.emit(f"Search of '{sift}' done after {elapsed}ms, {len(tags)} results")
        else:
            self._cacheTreeRestore()
            self._fetchAllMusic()

    @QtCore.pyqtSlot(str)
    def removeViewCache(self, folder):
        for path in dict(self.cacheView.db):
            if path.startswith(folder + "/"):
                del self.cacheView.db[path]
        self.cacheView.save()

    @QtCore.pyqtSlot()
    def saveTreeCache(self):
        self.cacheTree.db = {}
        for artist in range(self.parent.library.topLevelItemCount()):
            artist = self.parent.library.topLevelItem(artist)
            self.cacheTree.db[artist.text(1)] = []
            for album in range(artist.childCount()):
                album = artist.child(album)
                self.cacheTree.db[artist.text(1)].append(album.text(1))
        self.cacheTree.save()

    @QtCore.pyqtSlot(object)
    def selectEvent(self, selection):
        self.selection = selection
        QtCore.QMetaObject.invokeMethod(self.libraryView, "clear", QtCore.Qt.BlockingQueuedConnection)
        try:
            for item in selection:
                if item.text(1) == "__all__":
                    self._fetchAllMusic()
                    break
            else:
                paths = [i.text(1) for i in selection]
                self._fetchView(paths)
        except RuntimeError:
            pass




#!/usr/bin/python3
import os
import sys
from PyQt5 import QtGui, QtWidgets, QtCore, QtMultimedia, uic

try:
    from Xlib import X, display
    from Xlib.ext import record
    from Xlib.protocol import rq
except ImportError:
    pass

try:
    import qoob.backend.tools as tools
    import qoob.backend.player as player
    import qoob.backend.parser as parser
    import qoob.backend.library as library
    import qoob.frontend.preferences as preferences
    import qoob.frontend.trinkets as trinkets
    import qoob.frontend.playlist as playlist
    import qoob.ui.player
except ImportError:
    import backend.tools as tools
    import backend.player as player
    import backend.parser as parser
    import backend.library as library
    import frontend.preferences as preferences
    import frontend.trinkets as trinkets
    import frontend.playlist as playlist

LOCAL_DIR = os.path.dirname(os.path.realpath(__file__)) + "/"
DB_DIR = os.path.expanduser("~/.config/qoob/")


class Main(QtWidgets.QMainWindow):
    def __init__(self, parent):
        super().__init__()
        if "qoob.ui.player" in sys.modules:
            self.ui = qoob.ui.player.Ui_MainWindow()
            self.ui.setupUi(self)
        else:
            self.ui = uic.loadUi(LOCAL_DIR + "ui/player.ui", self)

        self.copied = []
        self.pending = False
        self.parent = parent
        self.logger = tools.Logger(name="qoob", path=DB_DIR)
        self.log = self.logger.new("qoob")
        self.preferences = preferences.PreferencesDatabase(self)
        self._styleInit()
        self.preferencesForm = preferences.PreferencesForm(self)
        self.popup = trinkets.PopupWidget(self)
        self.process = QtCore.QProcess()
        self.random = QtCore.QRandomGenerator()
        if not self.preferences.db["general"]["music database"]:
            self._findDefaultMusicFolder()

        self.player = player.MediaPlayer(self)
        self.parser = parser.Parser(self)
        self.tabWidget = playlist.Tabs(self)
        self.libraryView = self.tabWidget.tabs["Library viewer"]["playlist"]
        self.library = library.LibraryTree(self)
        self.trayIcon = trinkets.TrayIcon(self)

        self.player.currentMediaChanged.connect(self.currentMediaChanged)
        self.player.setItemColor.connect(self.setItemColor)
        self.player.setSliderRange.connect(self.ui.playbackSlider.setRange)
        self.player.setSliderValue.connect(self.ui.playbackSlider.setValue)
        self.player.setPlayPauseIcon.connect(self.ui.playButton.setIcon)
        self.player.setShuffleIcon.connect(self.ui.shuffleButton.setIcon)
        self.player.setRepeatIcon.connect(self.ui.repeatButton.setIcon)
        self.player.setStatusMessage.connect(self.setStatusMessage)
        self.player.shuffle = self.preferences.get("state", "shuffle")
        self.player.repeat = self.preferences.get("state", "repeat")
        self.ui.volumeSlider.setValue(self.player.volume())

        self.libraryThread = QtCore.QThread()
        self.libraryWorker = library.LibraryWorker(self)
        self.libraryWorker.moveToThread(self.libraryThread)
        self.libraryWorker.expand.connect(self.library.expand)
        self.libraryWorker.addArtist.connect(self.library.addArtist)
        self.libraryWorker.setStatusMessage.connect(self.setStatusMessage)
        self.libraryWorker.state.connect(self.libraryState)
        self.library.selectEvent.connect(self.librarySelectEvent)
        self.library.selectEvent.connect(self.libraryWorker.selectEvent)
        self.library.selectEvent.connect(self.player.clearViewerPlaylist)
        self.ui.libraryFilterLine.textEdited.connect(self.libraryFilterEditEvent)
        self.ui.libraryFilterLine.textEdited.connect(self.player.clearViewerPlaylist)
        self.ui.libraryFilterLine.keyPressEvent = self.libraryFilterKeyPressEvent
        self.libraryThread.start()

        self.tabWidget.keyEvent.connect(self.parseHotkeys)
        self.tabWidget.clearTabPlaylist.connect(self.player.clearTabPlaylist)
        self.preferencesForm.accepted.connect(self.preferencesSave)
        self.preferencesForm.resetMetadata.connect(self.libraryReset)
        self.preferencesForm.activateWindow()

        artEnabled = self.preferences.get("art", "show empty") and self.preferences.get("art", "enable")
        self.artLabel = trinkets.ArtLabel(self)
        self.artLabel.setArtButtonIcon.connect(self.ui.artButton.setIcon)
        self.artLabel.setFrameVisible.connect(self.ui.artFrame.setVisible)
        self.ui.horizontalLayout.addWidget(self.artLabel)
        self.ui.artFrame.setVisible(artEnabled)
        if self.preferences.get("art", "enable"):
            self.ui.artButton.setIcon(self.icon["art_on"])
        else:
            self.ui.artButton.setIcon(self.icon["art_off"])

        self.menu = playlist.Menu(self)
        self.menu.action.connect(self.action)

        self.parserThread = QtCore.QThread()
        self.parser.moveToThread(self.parserThread)
        self.parser.addArtist.connect(self.library.addArtist)
        self.parser.addAlbum.connect(self.library.addAlbum)
        self.parser.clear.connect(self.libraryClearEvent)
        self.parser.doneDb.connect(self.preferencesForm.dbDoneSlot)
        self.parser.startDb.connect(self.preferencesForm.dbStartSlot)
        self.parser.state.connect(self.libraryState)
        self.parser.filterEnable.connect(self.libraryFilterEnable)
        self.parser.setStatusMessage.connect(self.setStatusMessage)
        self.parser.saveTreeCache.connect(self.libraryWorker.saveTreeCache)
        self.parserThread.start()

        self.trayIcon.hideWindow.connect(self.hide)
        self.trayIcon.showWindow.connect(self.show)
        self.trayIcon.close.connect(self.close)
        self.trayIcon.play.connect(self.player.playPauseEvent)
        self.trayIcon.stop.connect(self.player.stopEvent)
        self.trayIcon.next.connect(self.player.nextEvent)
        self.trayIcon.previous.connect(self.player.previousEvent)
        self.trayIcon.shuffle.connect(self.player.shuffleEvent)
        self.trayIcon.showPreferencesForm.connect(self.preferencesForm.show)
        self.trayIcon.popup.connect(self.popup.display)

        if "Xlib" in sys.modules:
            self.mediaKeyMonitor = XMediaKeysMonitor()
            self.mediaKeyMonitor.playPause.connect(self.player.playPauseEvent)
            self.mediaKeyMonitor.stop.connect(self.player.stopEvent)
            self.mediaKeyMonitor.previous.connect(self.player.previousEvent)
            self.mediaKeyMonitor.next.connect(self.player.nextEvent)
            self.mediaKeyMonitor.pause.connect(self.player.pauseEvent)
            self.mediaKeyThread = QtCore.QThread()
            self.mediaKeyMonitor.moveToThread(self.mediaKeyThread)
            self.mediaKeyThread.started.connect(self.mediaKeyMonitor.init)
            self.mediaKeyThread.start()

        self.ui.statusRightLabel = QtWidgets.QLabel()
        self.ui.statusBar = QtWidgets.QStatusBar()
        self.ui.statusBar.addPermanentWidget(self.ui.statusRightLabel)
        self.ui.statusBar.setStyleSheet("QStatusBar::item { border: 0 }")
        self.ui.libraryLayout.insertWidget(1, self.library)
        self.ui.tabLayout.addWidget(self.tabWidget, 1, 1, 1, 8)

        self.ui.playButton.clicked.connect(self.player.playPauseEvent)
        self.ui.stopButton.clicked.connect(self.player.stopEvent)
        self.ui.nextButton.clicked.connect(lambda: self.player.nextEvent(forced=True))
        self.ui.previousButton.clicked.connect(self.player.previousEvent)
        self.ui.shuffleButton.clicked.connect(self.player.shuffleEvent)
        self.ui.repeatButton.clicked.connect(self.player.repeatEvent)
        self.ui.preferencesButton.clicked.connect(self.preferencesShow)
        self.ui.refreshButton.clicked.connect(self.libraryRefresh)
        self.ui.collapseButton.clicked.connect(self.library.collapseAll)
        self.ui.expandButton.clicked.connect(self.library.expandAll)
        self.ui.libraryFilterClearButton.clicked.connect(self.libraryFilterClear)
        self.ui.volumeSlider.valueChanged.connect(self.player.setVolume)
        self.ui.playbackSlider.installEventFilter(self)
        self.ui.volumeSlider.installEventFilter(self)
        self.ui.artButton.clicked.connect(self.artLabel.toggle)

        self._styleLoad()
        self.setStatusBar(self.ui.statusBar)
        self.setWindowTitle("qoob")
        self.setWindowIcon(self.icon["qoob"])
        width = self.preferences.get("state", "width")
        height = self.preferences.get("state", "height")
        self.resize(width, height)
        self.tabWidget.setFocus(True)
        self._resumePlayback()
        self.libraryWorker._fetchTree()
        self.show()

    def closeEvent(self, event):
        self.preferences.db["state"]["shuffle"] = self.player.shuffle
        self.preferences.db["state"]["repeat"] = self.player.repeat
        self.preferences.db["state"]["volume"] = self.player.volume()
        self.preferences.db["state"]["tab"] = self.tabWidget.tabBar().currentIndex()
        self.preferences.db["state"]["playback position"] = self.player.position()
        self.preferences.db["state"]["media state"] = self.player.state()
        self.preferences.db["state"]["file"] = self.player.path
        self.preferences.db["state"]["width"] = self.width()
        self.preferences.db["state"]["height"] = self.height()
        self.preferences.save()
        self.tabWidget.playlistSave()
        self.parent.exit()

    def eventFilter(self, obj, event):
        if event.type() == QtCore.QEvent.MouseButtonRelease:
            mouseEvent = QtGui.QMouseEvent(event)
            position = QtWidgets.QStyle.sliderValueFromPosition(obj.minimum(), obj.maximum(), mouseEvent.x(), obj.width())
            obj.setValue(position)
            if obj == self.ui.playbackSlider:
                self.player.setPosition(obj.value())
        return QtCore.QObject.event(obj, event)

    def changeEvent(self, event):
        # Override minimize event
        if event.type() == QtCore.QEvent.WindowStateChange:
            if self.windowState() & QtCore.Qt.WindowMinimized:
                if self.preferences.get("general", "tray minimize") and self.preferences.get("general", "tray icon"):
                    self.setWindowState(QtCore.Qt.WindowNoState)
                    self.hide() if self.isVisible() else self.show()

    def _delete(self, files):
        self.action("delete from playlist")
        for f in files:
            if os.path.isfile(f):
                os.remove(f)

            # Remove empty folder
            if self.preferences.get("general", "clean folder"):
                parent = os.path.abspath(os.path.join(f, os.pardir))
                if len(os.listdir(parent)) == 0:
                    os.rmdir(parent)
                    self.libraryRefresh()
                    return

        self.player.nextEvent()
        """
        # Remove item from library tree
        for item in range(self.library.topLevelItemCount()):
            item = self.library.topLevelItem(item)
            if self.library.deleteItem(item, f):
                break
        """

    def _deletePrompt(self, files):
        if self.preferences.get("general", "delete confirm"):
            names = ""
            for f in files:
                names += "\n" + os.path.basename(f)
            msg = QtWidgets.QMessageBox()
            msg.setWindowFlags(msg.windowFlags() | QtCore.Qt.WindowStaysOnTopHint)
            msg.setIcon(QtWidgets.QMessageBox.Warning)
            msg.setWindowTitle("Delete confirmation")
            msg.setText(f"Please confirm deletion of :\n{names}")
            msg.setStandardButtons(QtWidgets.QMessageBox.Apply | QtWidgets.QMessageBox.Cancel)
            if msg.exec_() == QtWidgets.QMessageBox.Apply:
                self._delete(files)
        else:
            self._delete(files)

    def _findDefaultMusicFolder(self):
        xdgDirectories = os.path.expanduser("~/.config/user-dirs.dirs")
        if os.path.isfile(xdgDirectories):
            with open(xdgDirectories) as f:
                for line in f.read().splitlines():
                    line = line.split("=")
                    if line[0] == "XDG_MUSIC_DIR":
                        musicFolder = line[1].replace("$HOME", "~")
                        musicFolder = os.path.expanduser(musicFolder[1:-1])
                        self.preferences.db["general"]["music database"].append(musicFolder)
                        self.preferences.save()
                        break

    def _styleInit(self):
        # Init icons
        iconTheme = self.preferences.get("general", "icon theme")
        iconPath = f"{LOCAL_DIR}icons/{iconTheme}/"
        self.icon = {}
        for icon in os.listdir(iconPath):
            iconName = os.path.splitext(icon)
            if iconName[1] == ".svg":
                self.icon[iconName[0]] = QtGui.QIcon(f"{iconPath}{icon}")

        # Init stylesheet
        stylesheet = f"QPushButton#libraryFilterClearButton {{ border-image: url({iconPath}filter_clear.svg); }}\n"
        stylesheet += f"QPushButton#libraryFilterClearButton:hover {{ border-image: url({iconPath}filter_hover.svg); }}"
        self.ui.libraryFilterClearButton.setStyleSheet(stylesheet)

    def _styleLoad(self):
        self.ui.playButton.setIcon(self.icon["play"])
        self.ui.stopButton.setIcon(self.icon["stop"])
        self.ui.nextButton.setIcon(self.icon["next"])
        self.ui.previousButton.setIcon(self.icon["previous"])
        self.ui.preferencesButton.setIcon(self.icon["preferences"])
        self.ui.refreshButton.setIcon(self.icon["refresh"])
        self.ui.collapseButton.setIcon(self.icon["collapse"])
        self.ui.expandButton.setIcon(self.icon["expand"])

        icon = {0: "repeat_off", 1: "repeat_all", 2: "repeat_single"}
        stateIcon = icon[self.player.repeat]
        self.ui.repeatButton.setIcon(self.icon[stateIcon])

        if self.player.shuffle:
            self.ui.shuffleButton.setIcon(self.icon["shuffle_on"])
        else:
            self.ui.shuffleButton.setIcon(self.icon["shuffle_off"])

        if self.preferences.get("art", "enable"):
            self.ui.artButton.setIcon(self.icon["art_on"])
        else:
            self.ui.artButton.setIcon(self.icon["art_off"])

    def _resumePlayback(self):
        self.player.resumePlayback = self.preferences.get("general", "resume playback")
        playing = self.preferences.get("state", "media state") == QtMultimedia.QMediaPlayer.PlayingState
        if self.player.resumePlayback and playing:
            self.player.setCurrentMedia(self.preferences.get("state", "file"))

    def _titleFormat(self, title):
        replace = ("artist", "album", "track", "title")
        for tag in replace:
            title = title.replace("%" + tag + "%", self.player.tags[tag])
        return title

    def _viewerLoadSort(self, content):
        self.tabWidget.tabBar().setCurrentIndex(0)
        tags = self.libraryWorker._fetchTag(content)
        tags = self.libraryWorker._sort(tags)
        self.libraryWorker._fillView(tags)

    def action(self, action):
        currentTab = self.tabWidget.current
        selection = currentTab.selectedItems()
        if action == "paste" and self.copied:
            for path in self.copied:
                item = currentTab.add(path)

            # Select pasted item
            currentTab.clearSelection()
            currentTab.setCurrentItem(item)

        elif selection:
            files = []
            for item in selection:
                files.append(item.text(5))

            if action == "copy":
                self.copied = files

            if action == "cut":
                self.copied = files
                for item in selection:
                    index = currentTab.indexOfTopLevelItem(item)
                    currentTab.takeTopLevelItem(index)

            elif action == "delete from playlist":
                for item in selection:
                    index = currentTab.indexOfTopLevelItem(item)
                    currentTab.takeTopLevelItem(index)

            elif action == "delete prompt":
                self._deletePrompt(files)

            elif action == "delete no confirm":
                self._delete(files)

            elif action == "browse":
                folder = os.path.abspath(os.path.join(currentTab.currentItem().text(5), os.pardir))
                cmd = f'{self.preferences.get("general", "file manager")} "{folder}"'
                self.process.startDetached(cmd)

    def currentMediaChanged(self, path):
        titlePopup = self._titleFormat(self.preferences.get("viewer", "notification format"))
        titleWindow = self._titleFormat(self.preferences.get("viewer", "title format"))
        titleTooltip = self._titleFormat(self.preferences.get("viewer", "tooltip format"))
        self.popup.display(titlePopup)
        self.setWindowTitle(titleWindow)
        self.trayIcon.setToolTip(titleTooltip)
        self.artLabel.load(path)
        if self.preferences.get("art", "media trigger"):
            self.artLabel.display()

    def libraryClearEvent(self):
        self.libraryWorker.sift = ""
        self.library.clear()
        self.ui.libraryFilterLine.clear()

    def libraryState(self, pending):
        self.pending = pending
        if pending:
            self.preferencesForm.setPending()
        else:
            self.library.addAllMusic()

    def libraryFilterClear(self):
        if self.ui.libraryFilterLine.text():
            self.ui.libraryFilterLine.clear()
            self.libraryFilterEditEvent()
        self.ui.libraryFilterLine.setFocus(True)

    def libraryFilterEnable(self, state):
        self.ui.refreshButton.setEnabled(state)
        self.ui.libraryFilterLine.setEnabled(state)
        self.ui.libraryFilterClearButton.setEnabled(state)

    def libraryFilterEditEvent(self, text=""):
        sift = QtCore.Q_ARG(str, text)
        QtCore.QMetaObject.invokeMethod(self.libraryWorker, "filterEvent", QtCore.Qt.QueuedConnection, sift)

    def libraryFilterKeyPressEvent(self, event):
        if event.key() == QtCore.Qt.Key_Escape:
            self.libraryFilterClear()
        QtWidgets.QLineEdit.keyPressEvent(self.ui.libraryFilterLine, event)

    def libraryRefresh(self):
        if not self.pending:
            QtCore.QMetaObject.invokeMethod(self.parser, "scanAll", QtCore.Qt.QueuedConnection)

    def libraryReset(self):
        if not self.pending:
            QtCore.QMetaObject.invokeMethod(self.parser, "reset", QtCore.Qt.QueuedConnection)

    def librarySelectEvent(self, selection):
        self.tabWidget.tabBar().setCurrentIndex(0)
        self.artLabel.load(selection[0].text(1))
        if self.preferences.get("art", "library trigger"):
            self.artLabel.display()

    def menuShow(self):
        if self.tabWidget.current.selectedItems() or self.copied:
            self.menu.popup(QtGui.QCursor.pos())

    def parseCommands(self, cmd):
        if "delete" in cmd:
            self.action("delete prompt")

        elif "delete-no-confirm" in cmd:
            self.action("delete no confirm")

        if "shuffle" in cmd:
            if len(cmd["shuffle"]) > 0:
                if cmd["shuffle"][0] == "on":
                    self.player.shuffleEvent(enable=True)
                elif cmd["shuffle"][0] == "off":
                    self.player.shuffleEvent(enable=False)
            else:
                self.player.shuffleEvent()
            self.popup.display("Shuffle enabled" if self.player.shuffle else "Shuffle disabled")

        if "repeat" in cmd:
            if len(cmd["repeat"]) > 0:
                states = {"off": 0, "all": 1, "single": 2}
                arg = cmd["repeat"][0]
                self.player.repeatEvent(state=states[arg])
            else:
                self.player.repeatEvent()
            states = {0: "off", 1: "all", 2: "single"}
            self.popup.display(f"Repeat {states[self.player.repeat]}")

        if "clear" in cmd:
            self.libraryView.clear()

        if "file" in cmd:
            if len(cmd["file"]) == 0:
                return False
            for f in cmd["file"]:
                if not os.path.isfile(f):
                    cmd["file"].remove(f)
            self._viewerLoadSort(cmd["file"])

        if "folder" in cmd:
            if len(cmd["folder"]) == 0:
                return False
            content = set()
            for folder in cmd["folder"]:
                if os.path.isdir(folder):
                    for root, subfolder, files in os.walk(folder):
                        for f in files:
                            path = os.path.join(root, f)
                            content.add(path)
            self._viewerLoadSort(content)

        if "stop" in cmd:
            self.player.stopEvent()

        if "previous" in cmd:
            self.player.previousEvent()

        if "next" in cmd:
            self.player.nextEvent()

        if "play" in cmd:
            self.player.playEvent()

        if "play-pause" in cmd:
            self.player.playPauseEvent()

        if "pause" in cmd:
            self.player.pauseEvent()

        if "quit" in cmd:
            self.close()

    def parseHotkeys(self, modifier, key):
        ctrl = (modifier == QtCore.Qt.ControlModifier)
        if ctrl and key == QtCore.Qt.Key_C:
            self.action("copy")
        elif ctrl and key == QtCore.Qt.Key_X:
            self.action("cut")
        elif ctrl and key == QtCore.Qt.Key_V:
            self.action("paste")
        elif key == QtCore.Qt.Key_Left:
            self.player.skip(-5)
        elif key == QtCore.Qt.Key_Right:
            self.player.skip(5)
        elif key == QtCore.Qt.Key_Backspace:
            self.action("delete from playlist")
        elif key == QtCore.Qt.Key_Delete:
            self.action("delete prompt")
        elif key == QtCore.Qt.Key_Space:
            self.player.playPauseEvent()

    def preferencesSave(self):
        self.preferencesForm.settingsSave()
        oldDb = self.preferences.get("general", "music database")
        newDb = self.preferencesForm.db["general"]["music database"]
        oldSort = self.preferences.get("viewer", "sorting routine")
        newSort = self.preferencesForm.db["viewer"]["sorting routine"]
        changedDb = not oldDb == newDb
        changedSortRoutine = not oldSort == newSort
        self.preferences.db = tools.copyDict(self.preferencesForm.db)
        self.preferences.save()

        if changedDb:
            for path in oldDb:
                if path not in newDb:
                    path = QtCore.Q_ARG(str, path)
                    QtCore.QMetaObject.invokeMethod(self.libraryWorker, "removeViewCache", QtCore.Qt.QueuedConnection, path)
                    QtCore.QMetaObject.invokeMethod(self.parser, "removeMetaDatabase", QtCore.Qt.QueuedConnection, path)
            self.libraryRefresh()

        if changedSortRoutine:
            self.libraryWorker.clearViewCache()

        self._styleInit()
        self._styleInit()
        self._styleLoad()
        self.artLabel.display()
        self.player.icon = self.icon
        self.artLabel.icon = self.icon
        self.tabWidget.initStyle()
        self.popup.updateView()
        if self.preferences.get("general", "tray icon"):
            self.trayIcon.show()
        else:
            self.trayIcon.hide()

    def preferencesShow(self):
        self.preferencesForm.show()
        self.preferencesForm.raise_()

    def setItemColor(self, color):
        current = self.tabWidget.current.currentItem()
        if current:
            self.tabWidget.current.currentItem().setColor(color)

    def setStatusMessage(self, left=None, right=None):
        if left:
            self.ui.statusBar.showMessage(left)
        if right:
            self.ui.statusRightLabel.setText(right)


class XMediaKeysMonitor(QtCore.QObject):
    playPause = QtCore.pyqtSignal()
    pause = QtCore.pyqtSignal()
    stop = QtCore.pyqtSignal()
    previous = QtCore.pyqtSignal()
    next = QtCore.pyqtSignal()

    def _begin(self):
        self.context = \
            self.dpy.record_create_context(0,
            [record.AllClients],
            [{
                'core_requests': (0, 0),
                'core_replies': (0, 0),
                'ext_requests': (0, 0, 0, 0),
                'ext_replies': (0, 0, 0, 0),
                'delivered_events': (0, 0),
                'errors': (0, 0),
                'client_started': False,
                'client_died': False,
                'device_events': (X.KeyPress, X.KeyRelease),
            }])
        self.dpy.record_enable_context(self.context, self._callback)
        self.dpy.record_free_context(self.context)

    def _callback(self, reply):
        if reply.category == record.FromServer:
            event, data = rq.EventField(None).parse_binary_value(reply.data, self.dpy.display, None, None)
            if event.type == X.KeyPress:
                key = self.dpy.keycode_to_keysym(event.detail, 0)
                self._parse(key)

    def _parse(self, key):
        if key == 269025044:
            self.playPause.emit()
        elif key == 269025045:
            self.stop.emit()
        elif key == 269025046:
            self.previous.emit()
        elif key == 269025047:
            self.next.emit()
        elif key == 269025073:
            self.pause.emit()

    @QtCore.pyqtSlot()
    def init(self):
        self.dpy = display.Display()
        if self.dpy.has_extension("RECORD"):
            self._begin()


def main(cmd=""):
    app = QtWidgets.QApplication([''])
    app.setQuitOnLastWindowClosed(False)
    gui = Main(app)
    gui.parseCommands(cmd)
    bus = tools.QDBusObject(parent=gui)
    sys.exit(app.exec_())


if __name__ == "__main__":
    main()
